******************************************************************************
*  No Copyright, released to the Public Domain
*
*  $Workfile:   y2kcmpcm.prg  $
*
*  Programmer.: The Ghost in the Machine
*  Date.......: 10/22/98
*  Description: Y2KCODE.EXE generated comments for Y2K remediation herein
*
*  *** PVCS revision log section - Do not modify! ***
*  $Log:   N:\vcs\source\exec\y2kcmpcm.prv  $
*
*     Rev 34.9   11 Nov 2000 16:21:32   GRLTC25
*  Got pickier about CTOD() conversions... actually MyCTOD()
*
*     Rev 34.8   01 Dec 1998 14:51:54   GRLTC25
*  Added SET EPOCH
*
*     Rev 34.7   01 Oct 2000 16:02:38   GRLTC25
*  Improved MAK file and #include's so that each PRG is more independent.
*  Also: Corrected sampling for text reports and delimited files.
*
*     Rev 34.6   27 Nov 1998 13:16:06   GRLTC25
*  Added additional date-range adjustment method "two year"
*
*     Rev 34.5   24 Nov 1998 10:39:48   GRLTC25
*  Added "fudge" factor for leap years other than 2000
*
*     Rev 34.4   21 Nov 1998 13:08:28   GRLTC25
*  Corrected some problems in SafeUse()
*
*     Rev 34.3   03 Nov 1998 09:55:30   GRLTC26
*  changes made for building with make file
*
*     Rev 34.2   26 Oct 1998 16:46:12   GRLTC25
*  Added endorsement boilerplate
*
*     Rev 34.1   26 Oct 1998 12:27:16   GRLTC25
*  Added date fudge to accomodate extra day aging around 2/28/2000
*
*     Rev 34.0   22 Oct 1998 11:44:14   GRLTC25
*  Initial revision.
*
******************************************************************************

#include "Y2KComp.ch"

MEMVAR gcAdjust

*----------------------------------------------------------------------------
FUNCTION Compare(cType, c1, c2, cGMBO, cCycle, cFreq, cName, oMeter, oSay)
*----------------------------------------------------------------------------
LOCAL nInc

/* Use an Epoch */
SET EPOCH TO 1950

/* Mention the version */
@0,0 Say "Dec 2, 1998 - 16:19"

/* Initialize the report and create header */
Report(_INIT)

Report("!     GMBO#:  "+cGMBO)
Report("!Test Cycle:  "+cCycle)
Report("!")
Report("! OUTPUT FILE COMPARISON PERFORMED TO THE FOLLOWING SPECIFICATION")
Report("!-----------------------------------------------------------------")
Report("! Comparison Type: "+cType)
Report("!   Sampling Rate: "+cFreq)
Report("!      First File: "+c1)
Report("!     Second File: "+c2)
Report("!-----------------------------------------------------------------")
Report("!")

/* Initialize the "difference aggregator" */
DiffAggr(_INIT)

/* Evaluate the sampling frequency value */
IF cFreq = NIL
    nInc := 1
ELSEIF cFreq = _FREQ_1
    nInc := 1
ELSEIF cFreq = _FREQ_3
    nInc := 3
ELSEIF cFreq = _FREQ_10
    nInc := 10
ELSEIF cFreq = _FREQ_50
    nInc := 50
ELSEIF cFreq = _FREQ_250
    nInc := 250
ELSEIF cFreq = _FREQ_1000
    nInc := 1000
ELSE
    nInc := 1
ENDIF

/* Evaluate the comparison type value */
IF cType = _TYPE_DBF
    DBFComp(c1, c2, nInc, oMeter, oSay)
ELSEIF cType = _TYPE_TXT
    DlmComp(c1, c2, nInc, " ", oMeter, oSay)
ELSEIF cType = _TYPE_DLM
    DlmComp(c1, c2, nInc, ",", oMeter, oSay)
ELSEIF cType = _TYPE_FLT
    MsgInfo("This comparison type is not currently available", "No comparison")
ELSE
    MsgInfo("This comparison type is not currently available", "No comparison")
ENDIF

/* Create summary results and activate the view of generated report */
Meter(_CLEAR)
DiffAggr(_FETCH)
Report("!")
Report("!-----------------------------------------------------------------")
Report("! ANY DIFFERENCE ENCOUNTERED ABOVE ARE SUFFICIENTLY EXPLAINED")
Report("! AS FOLLOWS:")
Report("!-----------------------------------------------------------------")
Report("!")
Report("!")
Report("!")
Report("!")
Report("!")
Report("!")
Report("!")
Report("!")
Report("!-----------------------------------------------------------------")
Report("! I CERTIFY THAT THIS TEST IS CONSIDERED A SUCCESSFUL COMPARISON.")
Report("!")
Report("!")
Report("!__________________________")
Report("! "+cName)

Results()
RETURN nil


*----------------------------------------------------------------------------
FUNCTION DlmComp(c1, c2, nInc, cDelim, oMeter, oSay)
*----------------------------------------------------------------------------
LOCAL i, j, nCnt := 0
LOCAL aRowCmp
LOCAL nStat
LOCAL aDiffs := {}
LOCAL nNorm
LOCAL fh1, fh2
LOCAL cLn1, cLn2
LOCAL cMess
LOCAL nLine := 0
LOCAL aCmp

/* First make sure we can use the files */
IF (-1==(fh1 := FOPEN(c1))) .OR. (-1==(fh2 := FOPEN(c2)))
    CLOSE ALL
    RETURN .F.
ENDIF

/* Do a little sanity check of file sizes */
IF (i := FILESIZE(fh1)) <> (j := FILESIZE(fh2))
    cMess := "Size of delimited-files is different ["+ALLTRIM(STR(i))+"/"     +;
             ALLTRIM(STR(j))+"]. This program is not smart enough "           +;
             "to reallign rows after gaps. Shall we proceed anyway?"
    IF .NOT. MsgYesNo(cMess, "Proceed?")
        CLOSE ALL
        RETURN .F.
    ENDIF
ENDIF

/* Try to get a modal date difference */
IF .NOT. DlmDateDiff(fh1, fh2, cDelim, @nNorm)
    CLOSE ALL
    RETURN .F.
ENDIF

/* Initialize the meter */
Meter(_TOTAL, FILESIZE(fh1), oMeter)
Meter(_CURRENT, 0, oMeter)

/* Compare the actual guts of the files */
FILETOP(fh1)
FILETOP(fh2)
DO WHILE ("" <> (cLn1 := FREADLN(fh1,1,255))) .AND.                            ;
         ("" <> (cLn2 := FREADLN(fh2,1,255)))

    aCmp := ArrayComp(Parse(cLn1, cDelim), Parse(cLn2, cDelim))
    IF .NOT. DiffAggr(_AGGR, aCmp, nNorm, ++nLine)
        Report("Processing has stopped because the maximum differences limit "+;
               "was exceeded--please see the difference summary below.")
        CLOSE ALL
        RETURN .F.
    ENDIF
    SetText( "Records compared: "+ ALLTRIM(STR(++nCnt)), oSay)
    Meter(_CURRENT, FSEEK(fh1,0,1), oMeter)
    SysRefresh()

    /* Maybe skip some lines */
    IF nInc > 1
        nLine += (nInc-1)
        FREADLN(fh1, nInc-1, 255)
        FREADLN(fh2, nInc-1, 255)
    ENDIF
ENDDO

SetText("", oSay)
Meter(_CURRENT, 0, oMeter)
CLOSE ALL
RETURN .T.


*----------------------------------------------------------------------------
FUNCTION DBFComp(c1, c2, nInc, oMeter, oSay)
*----------------------------------------------------------------------------
LOCAL i, j, nCnt := 0
LOCAL aRowCmp
LOCAL nStat
LOCAL aDiffs := {}
LOCAL nNorm
LOCAL aCmp

/* First make sure we can USE the DBFs */
IF .NOT. SafeUse(c1, "db1") .OR. .NOT. SafeUse(c2, "db2")
    CLOSE ALL
    RETURN .F.
ENDIF

/* Make sure the structures match (well enough) */
IF .NOT. StrctComp("db1", "db2")
    CLOSE ALL
    RETURN .F.
ENDIF

/* Do a little sanity check of row count */
IF .NOT. RowComp("db1", "db2")
    CLOSE ALL
    RETURN .F.
ENDIF

/* Try to get a modal date difference */
IF .NOT. DateDiff("db1", "db2", @nNorm)
    CLOSE ALL
    RETURN .F.
ENDIF

/* Initialize the meter */
Meter(_TOTAL, RECCOUNT(), oMeter)
Meter(_CURRENT, 0, oMeter)

/* Compare the actual guts of the DBFs */
SELECT ("db1")  ;  GO TOP
SELECT ("db2")  ;  GO TOP
DO WHILE .NOT. EOF()
    aCmp := ArrayComp(Scatter("db1"), Scatter("db2"))
    IF .NOT. DiffAggr(_AGGR, aCmp, nNorm, RECNO())
        Report("Processing has stopped because the maximum differences limit "+;
               "was exceeded--please see the difference summary below.")
        CLOSE ALL
        RETURN .F.
    ENDIF
    SetText( "Records compared: "+ ALLTRIM(STR(++nCnt)), oSay)
    Meter(_CURRENT, RECNO(), oMeter)
    SysRefresh()

    SELECT db2  ;  SKIP nInc
    SELECT db1  ;  SKIP nInc
ENDDO

SetText("", oSay)
Meter(_CURRENT, 0, oMeter)
CLOSE ALL
RETURN .T.


*----------------------------------------------------------------------------
FUNCTION DiffAggr(cWhat, aDiffs, nNorm, nRow)
*----------------------------------------------------------------------------
* Aggregate each difference array passed in here.
* Generate summary report in case the _FETCH argument is passed in.

LOCAL i
STATIC nRowDiff := 0
STATIC cRowDiff := ""
STATIC nOK      := 0
STATIC nBad     := 0
STATIC cItmDiff := ""
STATIC nNoNorm  := 0
STATIC cNoNorm  := ""
STATIC nDtFudge := 0

IF cWhat = _INIT           /* Initialize all the STATIC results */
    nRowdiff := 0
    cRowDiff := ""
    nOK      := 0
    nBad     := 0
    cItmDiff := ""
    nNoNorm  := 0
    cNoNorm  := ""
    nDtFudge := 0
    RETURN .T.

ELSEIF cWhat = _FUDGEDATE
    nDtFudge++

ELSEIF cWhat = _FETCH      /* Generate the diff summary report */

    Report(ALLTRIM(STR(nOK)) + " items were compared.")
    Report(ALLTRIM(STR(nBad)) + " differences were detected.")
    Report(ALLTRIM(STR(nDtFudge)) + " of the date comparisons were adjusted " +;
           "to account for the special data aging past leap days."             )

    Report(ALLTRIM(STR(nRowdiff))+" lines were different in length of item"   +;
           "list (i.e. they could not be parsed for comparison). "+cRowdiff    )

    Report(ALLTRIM(STR(nNoNorm)) + " dates were found to vary from the "      +;
           "expected difference. "+crlf+cNoNorm                                    )

    Report("The following non-date items were found to be non-identical "     +;
           "between the compared files: "+crlf+cItmDiff                        )

    RETURN .T.

ELSEIF LEN(aDiffs) = 0      /* Generate summarized lines diffs up to maximum */
    nBad++
    nRowdiff++
    cRowDiff += ALLTRIM(STR(nRow))+" : "
    RETURN (nRowdiff < _MAXROWDIFF)

ELSE                        /* Generate individual result totals and lists */
    FOR i = 1 TO LEN(aDiffs)
                            /* Logical comparison (not a date difference) */
        IF VALTYPE(aDiffs[i]) = "L"
            IF aDiffs[i]
                nOK++
            ELSE
                nBad++
                cItmDiff += "["+ALLTRIM(STR(nRow))+","+ALLTRIM(STR(i))+"] : "
            ENDIF

        ELSE                /* This is a date comparison in days */

            IF aDiffs[i] <> nNorm
                nBad++
                nNoNorm++
                cNoNorm += "["+ALLTRIM(STR(nRow))+","+ALLTRIM(STR(i))+"]"     +;
                           ALLTRIM(STR(aDiffs[i]))+"d : "
            ENDIF
        ENDIF
    NEXT i
ENDIF

RETURN (nBad < _MAXBAD)


*----------------------------------------------------------------------------
FUNCTION ArrayComp(a1, a2, nRow)
*----------------------------------------------------------------------------
LOCAL i
LOCAL aCmp

//-- Return an empty comparison list for different length arrays
IF LEN(a1) <> LEN(a2)
    RETURN {}
ELSE
    aCmp := ARRAY(LEN(a1))
ENDIF

//-- Return a list of either numeric date differences or T/F comparisons

/* Also, tally the count of the times where date substraction was adjusted
   for the special aging procedure around 2/28/00.  I.e. use the third param
   of .T. in SubtractDT()                                                     */

FOR i = 1 TO LEN(a1)
    IF EMPTY(a1[i]) .AND. EMPTY(a2[i])
        aCmp[i] := .T.
    ELSEIF VALTYPE(a1[i]) = "D" .AND. VALTYPE(a2[i]) = "D"
        aCmp[i] := SubtractDT(a1[i], a2[i], .T.)
    ELSEIF VALTYPE(a1[i]) = "D" .AND. VALTYPE(a2[i]) = "C"
        aCmp[i] := SubtractDT(a1[i], MyCTOD(a2[i]), .T.)
    ELSEIF VALTYPE(a1[i]) = "C" .AND. VALTYPE(a2[i]) = "D"
        aCmp[i] := SubtractDT(MyCTOD(a1[i]), a2[i], .T.)
    ELSE
        aCmp[i] := (a1[i] = a2[i])
    ENDIF
NEXT i
RETURN aCmp


*----------------------------------------------------------------------------
FUNCTION RowComp(cDb1, cDb2)
*----------------------------------------------------------------------------
LOCAL n, m, cMess
SELECT (cDB1)
n := RECCOUNT()
SELECT (cDB2)
m := RECCOUNT()
DBCOMMIT()
IF n <> m
    cMess := "Row count of DBFs is different ["+ALLTRIM(STR(n))+"/"    +;
             ALLTRIM(STR(m))+"]. This program is not smart enough "    +;
             "to reallign rows after gaps. Shall we proceed anyway?"
    RETURN MsgYesNo(cMess, "Proceed?")
ENDIF
RETURN .T.


*----------------------------------------------------------------------------
FUNCTION DateDiff(cDb1, cDb2, nNorm)
*----------------------------------------------------------------------------
LOCAL aDiffs := {}, aFreqs := {}, aOccur := {}, aNorm := {}
LOCAL aDts1 := {}, aDts2 := {}
LOCAL i := 0, j
LOCAL nFlds, nSamp
LOCAL d1, d2
LOCAL nPos, nChoice

SELECT (cDb2)  ;  GO TOP
SELECT (cDb1)  ;  GO TOP
nFlds := FCOUNT()

// First get raw difference list
DO WHILE i < 500 .AND. .NOT. EOF()
    FOR j = 1 TO nFlds
        SELECT (cDB1)
        d1 := FIELDGET(j)
        SELECT (cDB2)
        d2 := FIELDGET(j)

        // We are aiming for a list of (first 500) date differences
        IF VALTYPE(d1) = "D" .AND. VALTYPE(d2) = "D"  .AND.  ;
           !EMPTY(d1) .AND. !EMPTY(d2)
            i++
            AADD(aDiffs, SubtractDt(d1,d2,.F.)) // do not tally adjustments made
        ENDIF                                   // to date aging for leap day
    NEXT i

    SELECT (cDB1) ; SKIP
    SELECT (cDB2) ; SKIP
ENDDO
nSamp := i

// Next accumulate from list
ASORT(aDiffs)
FOR i = 1 TO LEN(aDiffs)
    nPos := ASCAN(aOccur, aDiffs[i])
    IF nPos = 0                   // For first occurance of a diff, add it to
        AADD(aOccur, aDiffs[i])   // the aOccur list, and add a freq value of 1
        AADD(aFreqs, 1)
    ELSE                          // A subsequent occurence of a diff,
        aFreqs[nPos]++            // just increment the frequency
    ENDIF
NEXT i

// Finally, appraise the distribution (and select a normalization, if possible)
IF LEN(aOccur) = 0
    nNorm := 0
    Report("No date values were found in the files being compared.")
ELSEIF LEN(aOccur) = 1
    nNorm := aOccur[1]
    Report("All sampled date differences between the DBFs were exactly "      +;
           ALLTRIM(STR(nNorm))+" days apart.  We'll use this as our norm."     )
ELSEIF LEN(aOccur) >= _TOOMANYDIFFS
    Report("Too many date differences were sampled to proceed meaningfully [" +;
           ALLTRIM(STR(LEN(aOccur)))+"--maybe a custom comparison is needed."  )
    RETURN .F.
ELSE
    FOR i = 1 TO LEN(aOccur)
        AADD(aNorm, STR(aOccur[i],6,0)+" days ["                              +;
                    ALLTRIM(STR((100*aFreqs[i]/nSamp),2,0))+"% of sample]"     )
    NEXT i
    nChoice := Choice("Several date differences were encountered. Please "    +;
                   "select the number-of-days normalization to use", aNorm     )
    IF nChoice = 0
        Report("The user cancelled the comparison because multiple date "     +;
               "differences were detected"                                     )
        RETURN .F.
    ELSE
        Report("The user selected a normalized date difference to be used "   +;
               "based on the sampled differences: "+ aNorm[nChoice]            )
        nNorm := aOccur[nChoice]
    ENDIF
ENDIF
RETURN .T.


*----------------------------------------------------------------------------
FUNCTION DlmDateDiff(fh1, fh2, cDelim, nNorm)
*----------------------------------------------------------------------------
LOCAL aDiffs := {}, aFreqs := {}, aOccur := {}, aNorm := {}
LOCAL aDts1 := {}, aDts2 := {}
LOCAL i := 0, j
LOCAL nSamp
LOCAL d1, d2
LOCAL cLn1, cLn2
LOCAL nPos, nChoice
LOCAL aLn1, aLn2

FILETOP(fh1)
FILETOP(fh2)

// First get raw difference list
DO WHILE i < 500 .AND. ("" <> (cLn1 := FREADLN(fh1,1,255)))                    ;
                 .AND. ("" <> (cLn2 := FREADLN(fh2,1,255)))

    aLn1 := Parse(cLn1, cDelim)
    aLn2 := Parse(cLn2, cDelim)

    /* Even if one line has more elements, we'll compare the initial ones */
    FOR j = 1 TO MIN(LEN(aLn1), LEN(aLn2))

        d1 := aLn1[j]
        d2 := aLn2[j]

        // We are aiming for a list of (first 500) date differences
        IF (VALTYPE(d1) = "D") .AND. (VALTYPE(d2) = "D")
            i++
            AADD(aDiffs, SubtractDt(d1,d2,.F.)) // do not tally adjustments made
        ENDIF                                   // to date aging for leap day
    NEXT i
ENDDO
nSamp := i

// Next accumulate from list
ASORT(aDiffs)
FOR i = 1 TO LEN(aDiffs)
    nPos := ASCAN(aOccur, aDiffs[i])
    IF nPos = 0                   // For first occurance of a diff, add it to
        AADD(aOccur, aDiffs[i])   // the aOccur list, and add a freq value of 1
        AADD(aFreqs, 1)
    ELSE                          // A subsequent occurence of a diff,
        aFreqs[nPos]++            // just increment the frequency
    ENDIF
NEXT i

// Finally, appraise the distribution (and select a normalization, if possible)
IF LEN(aOccur) = 0
    nNorm := 0
    Report("No date values were found in the files being compared.")
ELSEIF LEN(aOccur) = 1
    nNorm := aOccur[1]
    Report("All sampled date differences between the DBFs were exactly "      +;
           ALLTRIM(STR(nNorm))+" days apart.  We'll use this as our norm."     )
ELSEIF LEN(aOccur) >= _TOOMANYDIFFS
    Report("Too many date differences were sampled to proceed meaningfully [" +;
           ALLTRIM(STR(LEN(aOccur)))+"--maybe a custom comparison is needed."  )
    RETURN .F.
ELSE
    FOR i = 1 TO LEN(aOccur)
        AADD(aNorm, STR(aOccur[i],6,0)+" days ["                              +;
                    ALLTRIM(STR((100*aFreqs[i]/nSamp),2,0))+"% of sample]"     )
    NEXT i

    nChoice := Choice("Several date differences were encountered. Please "    +;
                   "select the number-of-days normalization to use", aNorm     )

    IF nChoice = 0
        Report("The user cancelled the comparison because multiple date "     +;
               "differences were detected"                                     )
        RETURN .F.
    ELSE
        Report("The user selected a normalized date difference to be used "   +;
               "based on the sampled differences: "+ aNorm[nChoice]            )
        nNorm := aOccur[nChoice]
    ENDIF
ENDIF
RETURN .T.


*----------------------------------------------------------------------------
FUNCTION SubtractDT(d1, d2, lTally)
*----------------------------------------------------------------------------
LOCAL nDiff, i
LOCAL aLeapDays := { CTOD("02/28/2000"),  ;
                     CTOD("02/28/1996"),  ;
                     CTOD("02/28/1992"),  ;
                     CTOD("02/28/1988"),  ;
                     CTOD("02/28/1984")   }

/* In fudging dates crossing the leap days, we will use strict inequality
   comparators.  The effect of this may produce odd effects if one of the
   days compared is strictly equal to 2/28/00.  I.e. crossing 2/28 by a day
   and including it will produce identical day substractions                  */

nDiff := d2 - d1

IF gcAdjust == _TWOYEAR     /* Specific assumption of two-year aging */
    IF nDiff == 731
        nDiff := 730
    ELSEIF nDiff == -731
        nDiff := -730
    ENDIF

ELSEIF gcAdjust == _LEAPDAY /* General assumpt that leap-day xing gets adjusted */

    FOR i = 1 TO LEN(aLeapDays)
        IF (d1 > aLeapDays[i] .AND. d2 < aLeapDays[i])

            nDiff++
            IIF(lTally, DiffAggr(_FUDGEDATE), NIL)

        ELSEIF (d2 > aLeapDays[i] .AND. d1 < aLeapDays[i])

            nDiff--
            IIF(lTally, DiffAggr(_FUDGEDATE), NIL)

        ENDIF
    NEXT i

ELSE                    /* If we are not performing leap-day adjustment */
    // do nothing       /* then there is no need for fancy conditions   */
ENDIF

RETURN nDiff

*----------------------------------------------------------------------------
FUNCTION Parse(cLn, cDelim)
*----------------------------------------------------------------------------
LOCAL aLn := {}
LOCAL i
LOCAL cFld := "", cChr := "", cLast

cLn := STRTRAN(cLn, crlf, "")

IF cDelim = SPACE(1)
    aLn := ParseWhite(cLn)
ELSE
    FOR i = 1 TO LEN(cLn)
        cLast := cChr
        IF cDelim <> (cChr := SUBSTR(cLn,i,1))
            cFld += cChr
        ELSEIF (cChr = cDelim) .AND. (cLast = cDelim)
            cFld := ""
            AADD(aLn, cFld)
        ELSE
            cFld := IIF(SUBSTR(cFld,1,1)='"', SUBSTR(cFld,2), cFld)
            cFld := IIF(SUBSTR(cFld,-1) ='"', SUBSTR(cFld,1,LEN(cFld)-1), cFld)
            cFld := IIF(EMPTY(MyCTOD(cFld)), cFld, MyCTOD(cFld))
            AADD(aLn, cFld)
            cFld := ""
        ENDIF
    NEXT i
    AADD(aLn,cFld)  /* Need to pick up final non-terminated field */
ENDIF
RETURN aLn


*----------------------------------------------------------------------------
FUNCTION ParseWhite(cLn)
*----------------------------------------------------------------------------
LOCAL aLn := {}
LOCAL nPos, i

DO WHILE LEN(cLn) > 0
    cLn := ALLTRIM(cLn)
    nPos := AT(SPACE(1), cLn)
    IF nPos = 0
        AADD(aLn, cLn)
        cLn := ""
    ELSE
        AADD(aLn, SUBSTR(cLn, 1, nPos-1))
        cLn := SUBSTR(cLn, nPos)
    ENDIF
ENDDO

FOR i = 1 TO LEN(aLn)
    aLn[i] := IIF(SUBSTR(aLn[i],1,1)='"', SUBSTR(aLn[i],2), aLn[i])
    aLn[i] := IIF(SUBSTR(aLn[i],-1) ='"', SUBSTR(aLn[i],1,LEN(aLn[i])-1), aLn[i])
    aLn[i] := IIF(EMPTY(MyCTOD(aLn[i])), aLn[i], MyCTOD(aLn[i]))
NEXT i

IF LEN(aLn) = 0
    AADD(aLn, "-- BLANK LINE --")
ENDIF

RETURN aLn


*----------------------------------------------------------------------------
FUNCTION StrctComp(cDb1, cDb2)
*----------------------------------------------------------------------------
LOCAL i
LOCAL aStrct1, aStrct2

SELECT (cDb1)
aStrct1 := DBSTRUCT()

SELECT (cDb2)
aStrct2 := DBSTRUCT()

IF LEN(aStrct1) <> LEN(aStrct2)
    Report("The field counts of the two DBFs differ--"+        ;
           "make sure the correct files were specified, "+     ;
           "and that a custom comparison type is not required" )
    RETURN .F.
ENDIF

FOR i = 1 TO LEN(aStrct1)
    IF aStrct1[i][1] <> aStrct2[i][1]
        Report("Field names in DBFs differ ["+aStrct1[i][1]+"/"+aStrct2[i][1]+ ;
               "]--make sure the correct files were specified, "+              ;
               "and that a custom comparison type is not required"             )
        RETURN .F.
    ELSEIF aStrct1[i][2] <> aStrct2[i][2]
        IF aStrct1[i][2] $ "CD" .AND. aStrct2[i][2] $ "CD"
            Report("A character-type date field appears to have been "+        ;
                   "remediated to hold a true date value ["+                   ;
                   aStrct1[i][1]+"/"+aStrct2[i][1]+"]"                         )

        ELSE
            Report("Field types in DBFs differ ["+aStrct1[i][1]+"/"+aStrct2[i][1]+;
                   "]--make sure the correct files were specified, "+          ;
                   "and that a custom comparison type is not required"         )
        RETURN .F.
        ENDIF
    ELSEIF aStrct1[i][3] <> aStrct2[i][3]
        Report("Field lengths in DBFs differ ["+aStrct1[i][1]+"/"+aStrct2[i][1]+;
               "]--make sure the correct files were specified, "+              ;
               "and that a custom comparison type is not required"             )
        RETURN .F.
    ELSEIF aStrct1[i][3] <> aStrct2[i][3]
        Report("Field decimals in DBFs differ ["+aStrct1[i][1]+"/"+aStrct2[i][1]+;
               "]--make sure the correct files were specified, "+              ;
               "and that a custom comparison type is not required"             )
        RETURN .F.
    ENDIF
NEXT i
RETURN .T.


*----------------------------------------------------------------------------
FUNCTION Report(cMess)
*----------------------------------------------------------------------------
LOCAL i, cSplit := ""
STATIC cRpt := ""

IF cMess = _INIT
    cRpt := ""
    SayScreen("Starting Comparison...")
ELSEIF cMess = _FETCH
    /* Don't actually change our report memo */
ELSEIF SUBSTR(cMess,1,1) = "!"
    cRpt += SUBSTR(cMess,2) + crlf
ELSE
    FOR i = 1 TO MLCOUNT(cMess, 65, 4, .T.)
        cSplit += MEMOLINE(cMess, 65, i, 4, .T.) + crlf
    NEXT i
    cRpt += cSplit + crlf
    SayScreen(cMess)
ENDIF
RETURN cRpt


*----------------------------------------------------------------------------
FUNCTION Scatter(cDB)
*----------------------------------------------------------------------------
LOCAL i, nFld
LOCAL aRow := {}

SELECT (cDB)
aRow := ARRAY(nFld := FCOUNT())
FOR i = 1 TO nFld
    aRow[i] := FIELDGET(i)
NEXT i
RETURN aRow


*----------------------------------------------------------------------------
FUNCTION SafeUse(cDB, cAlias)
*----------------------------------------------------------------------------
LOCAL nArea := SELECT(cAlias)

// Add extension if absent
cDB := IIF( SUBSTR(UPPER(cDB),-4)=".DBF", cDB, cDB+".DBF" )

IF .NOT. FILE(cDB)
    Report("File "+cDB+" cannot be located--check location and filename.")
    RETURN .F.
ENDIF

IF nArea = 0
    USE (cDB) NEW EXCLUSIVE ALIAS (cAlias)
    IF NETERR()
        Report("Problem opening "+cDb+" was encountered--check lock contention.")
        RETURN .F.
    ELSE
        GO TOP
    ENDIF
ELSE
    SELECT (nArea)
ENDIF
RETURN .T.

*----------------------------------------------------------------------------
FUNCTION MyCTOD(c)
*----------------------------------------------------------------------------
LOCAL dt := CTOD(c)
LOCAL y  := SPACE(4)
LOCAL m  := SPACE(2)
LOCAL d  := SPACE(2)

IF Disqualify(c)
    RETURN CTOD("  /  /  ")
ELSEIF .NOT. EMPTY(dt)
    RETURN dt
ELSEIF LEN(c) = 8                          .AND. ;
       VAL(y := SUBSTR(c,1,4)) > 1850      .AND. ;
       VAL(y) < 2050                       .AND. ;
       VAL(m := SUBSTR(c,5,2)) > 0         .AND. ;
       VAL(m) <= 12                        .AND. ;
       VAL(d := SUBSTR(c,7,2)) > 0         .AND. ;
       VAL(d) <= 31

    RETURN CTOD( m +"/"+ d +"/"+ y )
ENDIF
RETURN dt

*----------------------------------------------------------------------------
FUNCTION Disqualify(c)
*----------------------------------------------------------------------------
LOCAL i
FOR i = 1 TO LEN(c)
    IF UPPER(SUBSTR(c,i,1)) $ "ABCDEFGHIJKLMNOPQRSTUVWXYZ!"
        RETURN .T.
    ENDIF
NEXT i
RETURN .F.
